from ..interfaces._bsl_serial import _bsl_serial as bsl_serial
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type

import re, time

class SP_2150:  
    def __init__(self, device_sn="") -> None:
        self._target_device_sn = device_sn
        self.inst = inst.SP_2150
        self.device_id = ""
        self.logger = bsl_logger(self.inst)
        
        self.logger.info(f'Initiating bsl_instrument with target S/N="{device_sn}"...')
        n_try = 0
        while (n_try < 3):
            self._com = self._serial_connect()
            if self._com is not None:
                if self._com.serial_port is not None:
                    break
            self.logger.warning(f"{n_try} - Failed to connect, re-trying.....")
            n_try+=1
            time.sleep(2)
        if self._com is not None:
            if self._com.serial_port is not None:
                self.device_id = self._com.device_id
                self.logger.device_id = self._com.device_id
                self.__init_reset()
                self.logger.success(f"READY - Monochromator.\n\n\n")
                return
        self.logger.error(f"FAILED to connect Monochromator!\n\n\n")
        raise bsl_type.DeviceConnectionFailed

    def __del__(self, *args, **kwargs) -> None:
        self.close()
        return None

    def __init_reset(self) -> None:
        self.logger.info(f"Reseting Monochromator.....")
        self._com_cmd("MONO-RESET")
        self.logger.success(f"Monochronmator Power-On Reset Performed.")
        time.sleep(5)

    def _serial_connect(self) -> bsl_serial:
        try:
            com_port = bsl_serial(inst.SP_2150, self._target_device_sn)
        except Exception as e:
            self.logger.error(f"{type(e)}")
        return com_port

    def set_wavelength(self, wavelength) -> None:
        """
        - Set monochromator to the specified wavelength (in nm).

        Parameters
        -----------
        wavelength : `int`
            Wavelength (in nm) to set the monochromator to.
        """
        self._com_cmd(f"{wavelength} GOTO")
        self.logger.info(f"Output set to {wavelength} nm.")
        return None

    def get_wavelength(self) -> float:
        """
        - Get the current wavelength (in nm) of the monochromator
        with the format 250.0.

        Returns:
        -----------
        wavelength : `float`
            Current wavelength (in nm) of the monochromator.
        """
        wavelength = float(re.findall("\d+\.\d+",self._com_query("?NM"))[0])
        self.logger.info(f"Current operating point is at {wavelength} nm.")
        return wavelength

    def set_grating(self, grating) -> None:
        """
        - Select either the first or second grating. Requires approximately 20 seconds.
        Moves to the same wavelength as the previous grating or 200nm default if wavelength
        is not accessible by the selected grating.

        Parameters
        -----------
        grating : `int`
            Grating (1 or 2). First or second grating.
        """
        if (grating!=1 and grating!=2):
            self.logger.error("Grating must be 1 or 2")
            raise bsl_type.DeviceOperationError
        self._com_cmd(f"{grating} GRATING")
        self.logger.info(f"Grating set to {grating}.")
        return None

    def get_grating(self) -> int:
        """
        - Get the current grating position (1 or 2) of the monochromator.

        Returns:
        -----------
        grating : `int`
            Current grating position (1 or 2) of the monochromoator.
        """
        grating = int(re.findall("\d",self._com_query("?GRATING"))[0])
        self.logger.info(f"Current grating is {grating}.")
        return grating

    def _com_query(self, msg, timeout:float = 0.5) -> str:
        self._com.flush_read_buffer()
        self._com.write(msg+'\r')
        resp = self._com.readline()
        if "SERIAL" in resp:
            self._com.flush_read_buffer()
            self._com.write(msg+'\r')
            resp = self._com.readline()
        if resp in msg:
            resp = self._com.readline()
        return resp

    def _com_cmd(self, msg, timeout:float = 0.5) -> int:
        self._com.flush_read_buffer()
        self._com.write(msg+'\r')
        resp = self._com.readline()
        if "SERIAL" in resp:
            self._com.flush_read_buffer()
            self._com.write(msg+'\r')
            resp = self._com.readline()
        if "ok" not in resp:
            self.logger.error(f"message from device: \"{resp}\"")
            raise bsl_type.DeviceOperationError
        else:
            return 0

    def close(self) -> None:
        if self._com is not None:
            self._com.close()
            del self._com
        self.logger.success(f"CLOSED - Monochromator.\n\n\n")
        del self.logger
        pass




from ..interfaces._bsl_visa import _bsl_visa as bsl_visa
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type

import time, sys

class DC2200:
    def __init__(self, device_sn:str="") -> None:
        self.inst = inst.DC2200
        self.device_id=""
        self.logger = bsl_logger(self.inst)
        self.logger.info(f"Initiating bsl_instrument - DC2200({device_sn})...")
        if self._com_connect(device_sn):
            self.logger.device_id = self.device_id
            self.logger.success(f"READY - Thorlab DC2200 LED Controller \"{self.device_id}\"\".\n\n\n")
            self._reset_controller()
        else:
            self.logger.error(f"FAILED to connect to Thorlab DC2200 ({device_sn}) LED Controller!\n\n\n")
            raise bsl_type.DeviceConnectionFailed
        pass

    def __del__(self, *args, **kwargs) -> None:
        self.close()
        return None

    def _com_connect(self, device_sn:str) -> bool:
        try:
            self._com = bsl_visa(inst.DC2200, device_sn)
        except Exception as e:
            self.logger.error(f"{type(e)}")
            sys.exit(-1)
        if self._com is None:
            if self._com.com_port is None:
                return False
        self.device_id = self._com.device_id
        return True


    def _reset_controller(self) -> None:
        """
        - Performs a reset of the LED controller.
        """
        self._com.write("*RST")
        self.logger.info(f"LED Controller is Reset to initial states.")


    def get_screen_brightness(self) -> int:
        """
        - Get current on-boadr touch screen brightness in percentage.

        Returns
        --------
        brightness : `int`
            0 - 100 percent of current screen brightness.
        """
        brightness = int(float(self._com.query("DISPlay:BRIGhtness?"))*100)
        self.logger.info(f"Current screen brightness: {brightness}.")
        return brightness
    

    def set_screen_brightness(self, brightness:int) -> int:
        """
        - Set the on-boadr touch screen brightness in percentage from 0 to 100.

        Parameter
        --------
        brightness : `int`
            0 - 100 percent of current screen brightness.
        
        Returns
        --------
        brightness_readback : `int`
            0 - 100 percent of current screen brightness readback.
        """
        assert(brightness>=0 and brightness<=100)
        self._com.write("DISPlay:BRIGhtness %f" % (brightness/100))
        self.logger.info(f"Screen set to: {brightness}%.")
        brightness = int(float(self._com.query("DISPlay:BRIGhtness?"))*100)
        self.logger.info(f"Current screen brightness: {brightness}%.")
        return brightness
    

    def set_LED1_OFF(self) -> None:
        """
        - Turn off the output of LED1.
        """
        self._com.write("OUTPut1:STATe OFF")
        self.logger.info(f"LED1 state set to OFF.")


    def set_LED2_OFF(self) -> None:
        """
        - Turn off the output of LED1.
        """
        self._com.write("OUTPut2:STATe OFF")
        self.logger.info(f"LED2 state set to OFF.")


    def set_LED1_ON(self) -> None:
        """
        - Turn off the output of LED1.
        """
        self._com.write("OUTPut1:STATe ON")
        self.logger.info(f"LED1 state set to ON.")


    def set_LED2_ON(self) -> None:
        """
        - Turn off the output of LED1.
        """
        self._com.write("OUTPut2:STATe ON")
        self.logger.info(f"LED2 state set to ON.")


    def set_LED1_constant_current(self, current_mA:float) -> None:
        """
        - Set the LED1 to Constant Current mode with specified current setting.

        Parameter
        --------
        current_mA : `float`
            Desired current output in mA.
        """
        self.set_LED1_OFF()
        self._com.write("SOURce1:MODe CC")
        self.logger.info(f"LED1's mode set to Constant Current Mode.")
        self._com.write(f"SOURCE1:CCURENT:CURRENT {(current_mA/1000):.2f}")
        self.logger.info(f"LED1's output current set to {current_mA}mA")
        self.set_LED1_ON()


    def set_LED2_constant_current(self, current_mA:float) -> None:
        """
        - Set the LED1 to Constant Current mode with specified current setting.

        Parameter
        --------
        current_mA : `float`
            Desired current output in mA.
        """
        self.set_LED2_OFF()
        self._com.write("SOURce2:MODe CC")
        self.logger.info(f"LED2's mode set to Constant Current Mode.")
        self._com.write(f"SOURCE2:CCURENT:CURRENT {(current_mA/1000):.2f}")
        self.logger.info(f"LED2's output current set to {current_mA}mA")
        self.set_LED2_ON()

    def set_LED1_constant_brightness(self, percent:float) -> None:
        """
        - Set the LED1 to Constant Brightness mode with specified limit setting in 
            percentage from 0 to 100%.

        Parameter
        --------
        percent : `float`
            Desired brightness output in %Limit.
        """
        self.set_LED1_OFF()
        self._com.write("SOURce1:MODe CB")
        self.logger.info(f"LED1's mode set to Constant Brightness Mode.")
        self._com.write(f"SOURCE1:CBRightness:BRIGhtness {(percent):.2f}")
        self.logger.info(f"LED1's output brightness set to {percent}% of maximum limit.")
        self.set_LED1_ON()

    def set_LED2_constant_brightness(self, percent:float) -> None:
        """
        - Set the LED2 to Constant Brightness mode with specified limit setting in 
            percentage from 0 to 100%.

        Parameter
        --------
        percent : `float`
            Desired brightness output in %Limit.
        """
        self.set_LED2_OFF()
        self._com.write("SOURce2:MODe CB")
        self.logger.info(f"LED2's mode set to Constant Brightness Mode.")
        self._com.write(f"SOURCE2:CBRightness:BRIGhtness {(percent):.2f}")
        self.logger.info(f"LED2's output brightness set to {percent}% of maximum limit.")
        self.set_LED2_ON()

    def set_LED1_PWM(self, current_mA:float, frequency:float, duty_cycle:float, count:int=0) -> None:
        """
        - Set the LED1 to PWM mode with specified current in mA, switching frequency, duty_cycle in 
        percentage, and pulse counts.

        Parameter
        --------
        current_mA : `float`
            Desired current output in mA.

        frequency : `float`
            Desired PWM switching frequency in Hz.

        duty_cycle : 'float'
            Desired PWM Duty cycles.
        
        count : 'int' (default to 0)
            Desired pulse count, set to 0 for continuous operation
        """
        self.set_LED1_OFF()
        self._com.write("SOURce1:MODe PWM")
        self.logger.info(f"LED1's mode set to PWM Mode.")
        self._com.write(f"SOURCE1:PWM:CURRent {(current_mA/1000):.2f}")
        self.logger.info(f"LED1's output current set to {current_mA}mA.")
        self._com.write(f"SOURCE1:PWM:FREQency {frequency}.")
        self.logger.info(f"LED1's PWM frequency set to {frequency}Hz.")
        self._com.write(f"SOURCE1:PWM:DCYCle {duty_cycle}")
        self.logger.info(f"LED1's PWM Duty cycle set to {duty_cycle:.2f}%.")
        self._com.write(f"SOURCE1:PWM:COUNt {count}")
        self.logger.info(f"LED1's PWM pulse count set to {count}.")
        self.set_LED1_ON()

    def set_LED2_PWM(self, current_mA:float, frequency:float, duty_cycle:float, count:int=0) -> None:
        """
        - Set the LED2 to PWM mode with specified current in mA, switching frequency, duty_cycle in 
        percentage, and pulse counts.

        Parameter
        --------
        current_mA : `float`
            Desired current output in mA.

        frequency : `float`
            Desired PWM switching frequency in Hz.

        duty_cycle : 'float'
            Desired PWM Duty cycles.
        
        count : 'int' (default to 0)
            Desired pulse count, set to 0 for continuous operation
        """
        self.set_LED2_OFF()
        self._com.write("SOURce2:MODe PWM")
        self.logger.info(f"LED2's mode set to PWM Mode.")
        self._com.write(f"SOURCE2:PWM:CURRent {(current_mA/1000):.2f}")
        self.logger.info(f"LED2's output current set to {current_mA}mA.")
        self._com.write(f"SOURCE2:PWM:FREQency {frequency}.")
        self.logger.info(f"LED2's PWM frequency set to {frequency}Hz.")
        self._com.write(f"SOURCE1:PWM:DCYCle {duty_cycle}")
        self.logger.info(f"LED2's PWM Duty cycle set to {duty_cycle:.2f}%.")
        self._com.write(f"SOURCE2:PWM:COUNt {count}")
        self.logger.info(f"LED2's PWM pulse count set to {count}.")
        self.set_LED2_ON()

    # def get_LED_id(self) -> str:
    #     """
    #     - Get the sensor_id from the LED Controller.

    #     Returns
    #     --------
    #     sensor_id : `str`
    #         Sensor ID of the currently connected sensor.
    #     """
    #     sensor_id = self._com.query("SYST:SENS:IDN?").split(",")[0]
    #     self.logger.info(f"Current connected sensor: {sensor_id}.")
    #     return sensor_id


    def close(self) -> None:
        if self._com is not None:
            self._com.close()
            del self._com
        self.logger.info(f"CLOSED - Thorlab DC2200 LED Controller \"{self.device_id}\"\n\n\n")
        pass



from ..interfaces._bsl_visa import _bsl_visa as bsl_visa
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type
import time, sys

class PM100D:
    def __init__(self, device_sn:str="") -> None:
        self.inst = inst.PM100D
        self.device_id=""
        self.logger = bsl_logger(self.inst)
        self.logger.info(f"Initiating bsl_instrument - PM100D({device_sn})...")
        if self._com_connect(device_sn):
            self.logger.device_id = self.device_id
            self.run_update_power_meter()
            self.logger.success(f"READY - Thorlab PM100D Power Meter \"{self.device_id}\" with sensor \"{self.get_sensor_id()}\".\n\n\n")
        else:
            self.logger.error(f"FAILED to connect to Thorlab PM100D ({device_sn}) Power Meter!\n\n\n")
            raise bsl_type.DeviceConnectionFailed
        pass

    def __del__(self, *args, **kwargs) -> None:
        self.close()
        return None

    def _com_connect(self, device_sn:str) -> bool:
        try:
            self._com = bsl_visa(inst.PM100D, device_sn)
        except Exception as e:
            self.logger.error(f"{type(e)}")
            sys.exit(-1)
        if self._com is None:
            if self._com.com_port is None:
                return False
        self.device_id = self._com.device_id
        return True

    def run_update_power_meter(self) -> None:
        """
        - Performs an update to all relevent power meter parameters.

        - Enable info level loggging to see each query results.
        """
        self.get_preset_wavelength()
        self.get_attenuation_dB()
        self.get_average_count()
        self.get_measured_power()
        self.get_power_measuring_range()
        self.get_auto_range_status()
        self.get_measured_frequency()
        self.get_zero_magnitude()
        self.get_zero_state()
        self.get_photodiode_response()
        self.get_current_range()
        self.get_sensor_id()
        self.get_measured_current()
        pass

    def run_zero(self) -> None:
        """
        - Zero the power meter.
        """
        resp = self._com.write("SENS:CORR:COLL:ZERO:INIT")
        time.sleep(0.2)
        self.logger.info("Power Meter Zeroed.")
        return None

    def get_preset_wavelength(self) -> float:
        """
        - Get preset wavelength of interest for power measurement 
        from the power meter.

        Returns
        --------
        wavelength : `float`
            Preset wavelength of interest for power measurement.
        """
        try_count = 0
        while True:
            try:
                wavelength = float(self._com.query("SENS:CORR:WAV?"))
                self.logger.info( f"Current preset wavelenght: {repr(wavelength)}nm")
                break
            except:
                if try_count > 9:
                    self.logger.error("FAILED to acquire wavelength.")
                    raise bsl_type.DeviceOperationError
                else:
                    time.sleep(0.1)  #take a rest..
                    try_count = try_count + 1
                    self.logger.warning("timeout - Trying to get the wavelength again..")
        return wavelength
    
    def set_preset_wavelength(self, wl:float) -> float:
        """
        - Set preset wavelength of interest for power measurement 
        from the power meter.

        Parameters
        ----------
        wl : `float`
            Wavelength of interest for power measurement.

        Returns
        --------
        wavelength : `float`
            Preset wavelength of interest for power measurement readback
            from the power meter.
        """
        try_count = 0
        while True:
            try:
                self._com.write("SENS:CORR:WAV %f" % wl)
                time.sleep(0.005) # Sleep for 5 ms before rereading the wl.
                self.logger.info(f"Wavelength set to {wl:.1f}nm")
                break
            except:
                if try_count > 9:
                    self.logger.error( "Failed to set wavelength." )
                    raise bsl_type.DeviceOperationError
                else:
                    time.sleep(0.1)  #take a rest..
                    try_count = try_count + 1
                    self.logger.warning( "Timeout - trying to set wavelength again.." )

        return self.get_preset_wavelength()
    
    def get_attenuation_dB(self) -> float:
        """
        - Get current dB attenuation from the power meter.

        Returns
        --------
        att_dB : `float`
            Current dB attenuation of the power meter.
        """
        # in dB (range for 60db to -60db) gain or attenuation, default 0 dB
        attenuation_dB = float( self._com.query("SENS:CORR:LOSS:INP:MAGN?") )
        self.logger.info(f"Current attenuation at {attenuation_dB}dB.")
        return attenuation_dB

    def get_average_count(self) -> int:
        """
        - Get measurments count for each power measurement. 

        - Each measurement is approximately 3 ms.

        Returns
        --------
        count : `int`
            Number of measurements made for each power measurement,
            the result is the average of all measurements.
        """
        average_count = int( self._com.query("SENS:AVER:COUNt?") )
        self.logger.info( f"Current average count: {average_count}.")
        return average_count
    
    def set_average_count(self, cnt:int) -> int:
        """
        - Set measurments count for each power measurement,
        the final power measurement is the average of all 
        measured attempts. 

        - Each measurement is approximately 3 ms.

        parameter
        --------
        cnt : `int`
            Number of measurements made for each power measurement,
            the result is the average of all measurements.
        """
        self._com.write("SENS:AVER:COUNT %i" % cnt)
        self.logger.info(f"Average count is set to {cnt}.")
        return self.get_average_count()
            
    def get_measured_power(self) -> float:
        """
        - Get one power measurement form the power meter 
        with amount of `average_count` individual measurements.
        The final result is the average of all of theindividual 
        measurements.

        - Each measurement is approximately `3 * average_count`ms.

        Returns
        --------
        power : `float`
            The average of all of theindividual measurements.
        """
        power = float(self._com.query("MEAS:POW?"))
        self.logger.info(f"Current Power measured: {power*1000:.2f}mW.")
        return power
    
    def get_measured_power_avg(self, avg:int=1) -> float:
        """
        - Get one power measurement form the power meter 
        with amount of `average_count` individual measurements.
        The final result is the average of all of theindividual 
        measurements.

        - Each measurement is approximately `3 * average_count`ms.

        Returns
        --------
        power : `float`
            The average of all of theindividual measurements.
        """
        cnt = 0
        power = 0
        while cnt<avg:
            cnt+=1
            power += float(self._com.query("MEAS:POW?"))
        power = power/avg
        self.logger.info(f"Avg Power measured: {power*1000:.2f}mW.")
        return power
        
    #un tested
    def get_power_measuring_range(self) -> int:
        """
        - ???

        Returns
        --------
        range : `int`
            ???
        """
        power_range = float(self._com.query("SENS:POW:RANG:UPP?")) # CHECK RANGE
        self.logger.info(f"Power measuring range: {power_range*1000:.1f}mW.")
        return power_range

    #un tested
    def set_power_range(self, range:float) -> None:
        """
        - ???

        Parameters
        --------
        range : `float`
            ???
        """
        self._com.write("SENS:POW:RANG:UPP {}".format(range))
        self.logger.info(f"Set Power_measuring_range to {range}mW.")
        pass

    def get_auto_range_status(self) -> bool:
        """
        - Get the status of auto-ranging feature of the power meter.

        Returns
        --------
        auto-range : `bool`
            The status of auto-ranging feature of the power meter.
        """
        resp = self._com.query("SENS:POW:RANG:AUTO?")
        auto_range = bool(int(resp))
        self.logger.info(f"Current Auto_range status: {repr(auto_range)}.")
        return auto_range
    
    def set_auto_range(self, auto:bool = True) -> None:
        """
        - Set the status of auto-ranging feature of the power meter.

        Parameters
        --------
        auto : `bool`
            The requested status of auto-ranging feature of the power meter.
        """
        self.logger.info( f"Set Auto_range to: {repr(auto)}")
        if auto:
            self._com.write("SENS:POW:RANG:AUTO ON") # turn on auto range
        else:
            self._com.write("SENS:POW:RANG:AUTO OFF") # turn off auto range
    
    def get_measured_frequency(self) -> float:
        """
        - Get the measured frequency from the power meter.

        Returns
        --------
        freq : `float`
            Measured frequency `Hz` from the power meter.
        """
        frequency = float(self._com.query("MEAS:FREQ?"))
        self.logger.info(f"Measured frequency: {frequency:.1f}Hz.")
        return frequency

    def get_zero_magnitude(self) -> float:
        """
        - Get the zero_calibration magnitude (`Watts`) from the power meter.

        Returns
        --------
        zero_mag : `float`
            Current zero_calibration magnitude in `Watts`.
        """
        resp = self._com.query("SENS:CORR:COLL:ZERO:MAGN?")
        zero_magnitude = float(resp)
        self.logger.info(f"Current Zero_magnitude: {zero_magnitude*1000:.2f}mW.")
        return zero_magnitude
        
    def get_zero_state(self) -> bool: 
        """
        - Get the zero_calibration state from the power meter.

        Returns
        --------
        zero_state : `bool`
            Current zero_calibration state.
        """
        resp = self._com.query("SENS:CORR:COLL:ZERO:STAT?")
        zero_state = bool(int(resp))
        self.logger.info(f"Zero_state: {repr(zero_state)}.")
        return zero_state

    def get_photodiode_response(self) -> float:
        """
        - Get the photodiode_response magnitude (`A/W`) from the power meter.

        Returns
        --------
        photo_resp : `float`
            Current photodiode_response magnitude in `A/W`.
        """
        resp = self._com.query("SENS:CORR:POW:PDIOde:RESP?")
        #resp = self.ask("SENS:CORR:VOLT:RANG?")
        #resp = self.ask("SENS:CURR:RANG?")
        photodiode_response = float(resp) # A/W
        self.logger.info(f"Current Photodiode_response: {photodiode_response*1000:.1f}mA/W.")
        return photodiode_response 
    
    def get_measured_current(self) -> float:
        """
        - Get the measured_current magnitude (`A`) from the power meter.

        Returns
        --------
        current : `float`
            Current measured_current magnitude in `Amps`.
        """
        resp = self._com.query("MEAS:CURR?")
        current = float(resp)
        self.logger.info(f"Measured current: {current*1000:.1f}mA.")
        return current
    
    def get_current_range(self) -> float:
        """
        - Get the current_range upper bound (`A`) from the power meter.

        Returns
        --------
        max_current : `float`
            Current max current_range magnitude in `Amps`.
        """
        resp = self._com.query("SENS:CURR:RANG:UPP?")
        current_range = float(resp)
        self.logger.info(f"Preset current_range: {current_range*1000:.1f}mA.")
        return current_range

    def get_sensor_id(self) -> str:
        """
        - Get the sensor_id from the power meter.

        Returns
        --------
        sensor_id : `str`
            Sensor ID of the currently connected sensor.
        """
        sensor_id = self._com.query("SYST:SENS:IDN?").split(",")[0]
        self.logger.info(f"Current connected sensor: {sensor_id}.")
        return sensor_id


    def close(self) -> None:
        if self._com is not None:
            self._com.close()
            del self._com
        self.logger.info(f"CLOSED - Thorlab PM100D Power Meter \"{self.device_id}\"\n\n\n")
        pass



from pathlib import Path
import numpy as np
from .mantis_file import mantis_file
from numpy.typing import NDArray
from tqdm import tqdm
from loguru import logger
import re, os

class mantis_folder:
    def __init__(self, path:Path, sort_with_exp:bool=False, x3_conv:bool=False, x3_conv_param=0.48):
        """
        Load objects points for a folder of mantisCam recording files.

        Parameters
        ----------
        sort_with_exp : `bool`
            Sort all video files with the exposure time encoded in its file name.
            File name MUST contains exp time info in the format of "x.xms"

        Returns
        -------
        mantis_folder : `mantis_folder`
            A object with all information and data of mantisCam recoridings in the folder.
        """
        if type(path) is str:
            path = Path(path)
        self.path = path
        self.x3_conv = x3_conv
        self.x3_conv_param = x3_conv_param
        self.__videos = dict()
        self.__init_mantis_video_dict(sort_with_exp)
        
    def __extract_time_key(self, file_name):
        match = re.search(r'\d+\.\d+(?=ms)', file_name)
        return float(match.group())  # Extract the float value from the time string without the 'ms' unit

    def __init_mantis_video_dict(self, sort_with_exp:bool=True):
        filepaths = [f for f in self.path.iterdir() if f.is_file() and f.suffix == '.h5']
        for filepath in filepaths:
            self.__videos[filepath.name] = mantis_file(filepath, x3_conv=self.x3_conv, conv_param=self.x3_conv_param)
        if(sort_with_exp):
            self.__videos = dict(sorted(self.__videos.items(), key=lambda item: self.__extract_time_key(item[0])))

    def __getitem__(self,name:str) -> mantis_file:
        return self.__videos[name]

    @property
    def n_videos(self):
        return len(self.__videos)

    @property
    def name_videos(self):
        filenames = [f.name for f in self.path.iterdir() if f.is_file() and f.suffix == '.h5']
        sorted_filenames = sorted(filenames, key=lambda f: os.stat(os.path.join(self.path, f)).st_mtime)
        return sorted_filenames
    
    @property
    def arr_files(self) -> list[mantis_file]:
        """
        Return all the files in the mantisFolder object.

        Returns
        -------
        arr_videos : `numpy.ndarray`
            An array of all the video frames in all the mantisFiles.
            in the shape of [num_files]
        """
        return np.array(list(self.__videos.values()))
    
    @property
    def arr_videos(self) -> NDArray[np.uint16]:
        """
        Return all the frames in all of the files in the mantisFolder object.

        Returns
        -------
        arr_videos : `numpy.ndarray`
            An array of all the video frames in all the mantisFiles.
            in the shape of [num_files, n_row, n_col, n_channels]
        """
        file_shape = self.arr_files[0].raw_data_shape
        arr_size = self.n_videos *file_shape[0]*file_shape[1]*file_shape[2]*file_shape[3]*2/1000/1000/1000
        logger.warning(f"Your are trying to load the entire folder into a single dataset, this may crush your python kernel if your RAM is not enough! \nEstimate RAM Requirement: {arr_size:.2f}GB")
        return np.array([file.frames for file in tqdm(self.__videos.values(), desc="Loading Dataset")])
    

    def find_key(self, key:str) -> str:
        """
        Find the key of a file in the folder.

        Parameters
        ----------
        key : `str`
            The key to search for.

        Returns
        -------
        key : `str`
            The key of the file.
        """
        for file_name in self.name_videos:
            if key in file_name:
                return file_name
        raise ValueError(f"Key {key} not found in folder {self.path}")


from typing import Any
from loguru import logger
from datetime import datetime, timezone

from ..interfaces._bsl_serial import _bsl_serial as bsl_serial
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type

try:
    from MantisCam import Messenger
    import zmq
except ImportError:
    pass
import enum, sys, os, atexit, time
logger_opt = logger.opt(ansi=True)

class MantisCamCtrl:
    TIMEOUT_SEC = 120
    __ZMQ_CMD_DELAY = 0.01

    class FILE_SAVING_MODE(enum.Enum):
        CUSTOM = "Custom"; TIMSSTAMP = "Timestamp"

    def __init__(self, device_sn="", port_cmd_pub: int=60000, port_cmd_sub: int=60001, port_vid_sub: int=60011, *, log_level: str = "TRACE", is_GSENSE: bool = False):
        logger_opt.info(f"Initiating bsl_instrument - MantisCam({device_sn})...")
        url_prefix = 'tcp://127.0.0.1:'

        self.url_cmd_pub = url_prefix + str(port_cmd_pub)
        self.url_cmd_sub = url_prefix + str(port_cmd_sub)
        self.url_vid_sub = url_prefix + str(port_vid_sub)
        
        self.log_level = log_level
        self.is_recording = False
        self.__e_exp_time = 50
        self.__e_exp_matched = False
        self.__is_GSENSE = is_GSENSE

        self.__logger_init()
        self.__zmq_init()
        time.sleep(1)
        logger.success(f"Camera - ZMQ interface connected, camera initlized!")
        self.set_file_name(time_stamp_only=True)
        return


    def __zmq_init(self):
        # Initialize ZMQ for receiving video metadata and command, and send commands
        self.ctx = zmq.Context()
        self.cmd = Messenger(self.ctx, self.url_cmd_pub, self.url_cmd_sub, 'cmd', '')
        self.vid = Messenger(self.ctx, None, self.url_vid_sub, 'vid', 'raw')
        # We register ther termination function so we don't have to explicitly call it upon normal and abnormal termination
        atexit.register(self.__zmq_term)
        self.poller = zmq.Poller()
        self.poller.register(self.cmd.skt_sub, zmq.POLLIN)
        self.poller.register(self.vid.skt_sub, zmq.POLLIN)
        self.__zmq_send('cam', 'exp-00', dict([('exp-00', self.__e_exp_time)]))
        self.__zmq_send('widget', 'exp-00', dict([('exp-00', self.__e_exp_time)]))
        logger.success(f"Command PUB [{self.url_cmd_pub}] SUB [{self.url_cmd_sub}]")
        logger.success(f"Video SUB [{self.url_vid_sub}]")


    def __zmq_term(self):
        self.cmd.close()
        self.vid.close()
        self.ctx.term()
        logger.trace(f"ZMQ terminated.")


    def __zmq_recv(self):
        skts = dict(self.poller.poll(timeout=0))
        if self.cmd.skt_sub in skts:
            topic, name, msg = self.cmd.recv()
            if topic == 'file':
                if name == 'recording_status':
                    self.is_recording = msg['recording']
                        
        if self.vid.skt_sub in skts:
            # We always need to peak frame, because otherwise it will queue up in our receiving end
            topic, name, msg, frame_dump = self.vid.peak_frame()
            # If user hit run button (self.procedure_run), and we are waiting for exposure to match (that's why we need
            #   to read frame metadata at first place: see if the incoming frame exposure match our set exposure), and
            #   we are not waiting for complete. This limits the checking phase to the Smart mode, and only the period
            #   after set exposure and before start recording
            if topic == 'raw':
                if 'frame_meta' in msg:
                    if 'int-set' in msg['frame_meta']:
                        received_exposure_ms = round(float(msg['frame_meta']['int-set']), 0)
                        #logger.trace(f"Received frame exposure {received_exposure_ms}, expected frame exposure {self.__e_exp_time}.")
                        # If match, we procede to the next phase: recording
                        if received_exposure_ms == self.__e_exp_time:
                            self.__e_exp_matched = True
                        if self.__is_GSENSE:
                            self.__e_exp_matched = True
        time.sleep(self.__ZMQ_CMD_DELAY)


    def __zmq_send(self, topic: str, name: str, msg: Any):
        # Abstracted zmq command send with logging
        logger.trace(f"Sending t:n:m={topic}:{name}:{msg}")
        self.cmd.send(name, msg, topic=topic)
        time.sleep(self.__ZMQ_CMD_DELAY)


    def __logger_init(self):
        home_dir = os.path.expanduser('~')
        # Log to file and terminal at the same time
        logger.remove()
        logger.add(os.path.join(home_dir, 'MantisCam', 'AutoRecord', 'logs', datetime.now(timezone.utc).strftime('%Y-%m-%d_%H-%M-%S_%f')[:-3]+'.log'), 
                        enqueue=True, level=0, backtrace=True, diagnose=True,
                        format="{time:YYYY-MM-DD HH:mm:ss.SSSZZ} | {process: <5} | {level: <8} | {file}:{function}:{line} > {message}")
        logger.add(sys.stdout, 
                        enqueue=True, level=self.log_level, backtrace=True, diagnose=True,
                        format="<green>{time:YYYY-MM-DD HH:mm:ss.SSSZZ}</green> | <magenta>{process: <5}</magenta> | <level>{level: <8}</level> | <blue>{file}:{function}:{line}</blue> > <level>{message}</level>")
        atexit.register(self.__logger_term)


    def __set_frames_per_file(self, frames: int):
        # Send frames per file value to the file saving process
        self.__zmq_send('file', 'frames_per_file', dict(frames_per_file=frames))
    

    def set_exposure_ms(self, exp_time_ms: float) -> None:
        """
        - Set the camera exposure time to desired value in milliseconds.

        Parameters
        ----------
        exp_time_ms : `int`
            Desired exposure time in milliseconds.
        """

        # Helper function to set camera and GUI exposure values
        self.__zmq_send('cam', 'exp-00', dict([('exp-00', exp_time_ms)]))
        self.__zmq_send('widget', 'exp-00', dict([('exp-00', exp_time_ms)]))
        self.__e_exp_time = exp_time_ms
        self.__e_exp_matched = False
        time_start = time.time()
        while True:
            self.__zmq_recv()
            if self.__e_exp_matched:
                if self.__is_GSENSE:
                    time.sleep(1)
                logger.info(f"Camera - Exposure time set to {self.__e_exp_time}ms and verified.")
                break
            if time.time() - time_start > self.TIMEOUT_SEC:
                logger.error(f"ERROR - Camera exposure time not match, timed out!")
                raise bsl_type.DeviceTimeOutError


    def set_file_name(self, file_name: str="video", time_stamp_only:bool=False) -> None:
        """
        - Set the file name to be saved. 

        Parameters
        ----------
        file_name : `str`
            (ONLY used when time_stamp_only is set to False)
            Desired saving file name.
            Warning: If filename already existed in the saving directory, following files will have a 
            timestamp attached to the end.

        time_stamp_only : 'bool'
            (default to False)
            Enable timestamp only file name mode. No file_name string need to be provided then.
        """
        if time_stamp_only:
            self.__zmq_send('file', 'file_name', dict([('mode', 'Timestamp')]))
            logger.info(f'Camera recording filename changed to Timestamp only mode')
        else:
            self.__zmq_send('file', 'file_name', dict([('mode', 'Custom'), ('name', file_name)]))
            logger.info(f'Camera recording filename changed to {file_name}')

    
    def set_folder_name(self, create_new_folder:bool=True, time_stamp_only:bool=False, folder_name: set="video"):
        """
        - Set the file name to be saved. 

        Parameters
        ----------
        create_new_folder : 'bool'
            (default to True)
            If set to False, all recording files will be saved to the ROOT folder of the saving path.

        time_stamp_only : 'bool'
            (default to False)
            Enable timestamp only folder_name mode. No folder_name string need to be provided then.

        folder_name : `str`
            (ONLY used when 'time_stamp_only' is set to False AND 'create_new_folder' is set to True)
            Desired saving folder name.
        """
        if create_new_folder:
            if time_stamp_only:
                self.__zmq_send('file', 'folder_name', dict([('mode', 'Timestamp')]))
                logger.info(f'Camera recording folder name changed to Timestamp only mode.')
            else:
                self.__zmq_send('file', 'folder_name', dict([('mode', 'Custom'), ('name', folder_name)]))
                logger.info(f'Camera recording folder name changed to {folder_name}.')
        else:
            self.__zmq_send('file', 'folder_name', dict([('mode', 'Do Not Create New Folder')]))
            logger.info(f"No new folder will be created for the recording files.")



    def set_start_record(self, n_frames: int = 100, wait_until_done: bool = True):
        """
        - Start recording with desired frame counts per file, optional blocking/non-blocking recording
        that is default to blocking recording.

        Parameters
        ----------
        n_frames : `int`
            (default to 100)
            Desired number of frame per file.

        wait_until_done : `bool`
            (default to True)
            Blocking/non-blocking recording select.
            If True: this function will halt until the recording is done and camera is ready again.
            If False: this function will return as soon as camera started recording.
        """
        logger.info(f"Camera - Start recording...")
        # Set the frame per file
        self.__set_frames_per_file(n_frames)
        # Set start recording to GUI and file saving process
        self.__zmq_send('file', 'record', dict(record=True))
        time_start = time.time()
        while True:
            self.__zmq_recv()
            if self.is_recording:
                break
            if time.time() - time_start > 5:
                logger.error(f"ERROR - Unable to start recording, timed out!")
                raise bsl_type.DeviceTimeOutError
        logger.info(f"Camera - Recording with {n_frames} frames per file and exposure time of {self.__e_exp_time}ms.")
        if wait_until_done:
            time_start = time.time()
            while True:
                self.__zmq_recv()
                if not self.is_recording:
                    break
                if time.time() - time_start > (self.__e_exp_time * n_frames)/500 + 20:
                    logger.error(f"ERROR - Unable to finish recording, timed out! timeout @ {time.time() - time_start:.0f} seconds.")
                    # raise bsl_type.DeviceTimeOutError
        logger.info("Camera - Recording Finished!")
        


    def __logger_term(self):
        logger.info("Auto-Recording Utility terminated.")
        logger.complete()


from loguru import logger
from ..headers._bsl_inst_info import _bsl_inst_info_list
from ..headers import _bsl_type
import re
try:
    import pyvisa as pyvisa
except ImportError:
    pass
logger_opt = logger.opt(ansi=True)

# @logger_opt.catch
class _bsl_visa:

    def __init__(self, target_inst:_bsl_inst_info_list, device_sn:str="") -> None:
        #Init logger_opt by inherit from parent process or using a new one if no parent logger_opt
        logger_opt.info("    Initiating bsl_visa_service...")
        self.visa_resource_manager = pyvisa.ResourceManager()

        self.inst = target_inst
        self.target_device_sn = device_sn
        self._connect_visa_device()
        if self.com_port is None:
            logger_opt.error(f"<light-blue><italic>{self.inst.MODEL} ({self.target_device_sn})</italic></light-blue> not found on VISA/SCPI ports.")
        pass

    def __del__(self) -> None:
        self.close()

    def _find_device_vpid(self) -> None:
        resource_list = self.visa_resource_manager.list_resources()
        logger.debug(f"    bsl_VISA - Currently opened devices: {repr(self.visa_resource_manager.list_opened_resources())}")
        for port in resource_list:
            logger_opt.debug(f"    Found bus device <light-blue><italic>{port}</italic></light-blue>")
            if port in str(self.visa_resource_manager.list_opened_resources()):
                logger_opt.warning(f"    BUSY - Device <light-blue><italic>{port}</italic></light-blue> is busy, moving to next available device...")
                continue
            if (self.inst.USB_PID in port) and (self.inst.USB_VID in port):
                logger_opt.debug(f"    {self.inst.MODEL} is found with USB_PID/VID search.")
                temp_com_port = self.visa_resource_manager.open_resource(port)
                re_result = re.search(self.inst.SN_REG, temp_com_port.query(self.inst.QUERY_CMD).strip())
                if re_result is not None:
                    device_id = re_result.group(0)
                else:
                    device_id = "UNABLE_TO_OBTAIN"
                if self.target_device_sn not in device_id:
                    temp_com_port.close()
                    logger_opt.warning(f"    S/N Mismatch - Device <light-blue><italic>{port}</italic></light-blue> with S/N <light-blue><italic>{device_id}</italic></light-blue> found, not <light-blue><italic>{self.target_device_sn}</italic></light-blue> as requested, moving to next available device...")
                    continue
                temp_com_port.close()
                return port
            if ( str(int(self.inst.USB_PID,16)) in port and str(int(self.inst.USB_VID,16)) in port):
                logger_opt.debug(f"    {self.inst.MODEL} is found with USB_PID/VID search.")
                temp_com_port = self.visa_resource_manager.open_resource(port)
                device_id = re.search(self.inst.SN_REG, temp_com_port.query(self.inst.QUERY_CMD).strip()).group(1)
                if self.target_device_sn not in device_id:
                    temp_com_port.close()
                    logger_opt.warning(f"    S/N Mismatch - Device <light-blue><italic>{port}</italic></light-blue> with S/N <light-blue><italic>{device_id}</italic></light-blue> found, not <light-blue><italic>{self.target_device_sn}</italic></light-blue> as requested, moving to next available device...")
                    continue
                temp_com_port.close()
                return port
        return None

    def _connect_visa_device(self) -> None:
        port = self._find_device_vpid()
        self.com_port = None
        if port is not None:
            self.com_port = self.visa_resource_manager.open_resource(port)
        if self.com_port is not None:
            self.device_id = self.query(self.inst.QUERY_CMD).strip()
            if self.inst.QUERY_E_RESP not in self.device_id:
                logger_opt.error(f"    FAILED - Wrong device identifier (E_RESP) is returned!")
                raise _bsl_type.DeviceConnectionFailed
            self.device_id = re.search(self.inst.SN_REG, self.device_id).group(0)
            logger_opt.success(f"    {self.inst.MODEL} with DEVICE_ID: <light-blue><italic>{self.device_id}</italic></light-blue> found and connected!")
        pass

    def query(self, cmd:str):
        logger_opt.trace(f"        {self.inst.MODEL} - com-VISA - Query to {self.inst.MODEL} with {cmd}")
        resp = self.com_port.query(cmd).strip()
        logger_opt.trace(f"        {self.inst.MODEL} - com-VISA - Resp from {self.inst.MODEL} with {repr(resp)}")
        return resp
    
    def write(self, cmd:str) -> None:
        logger_opt.trace(f"        {self.inst.MODEL} - com-VISA - Write to {self.inst.MODEL} with {cmd}")
        self.com_port.write(cmd)
        pass

    def set_timeout_ms(self, timeout:int) -> None:
        self.com_port.timeout = timeout
        pass

    def close(self) -> None:
        if self.com_port is not None:
            self.com_port.close()
        pass


        


class _bsl_type:
    class CustomError(Exception):
        pass
    class DeviceConnectionFailed(CustomError):
        pass
    class DeviceOperationError(CustomError):
        pass
    class DeviceTimeOutError(CustomError):
        pass
    class DeviceInconsistentError(CustomError):
        pass


import imp
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type
import numpy
import seabreeze.spectrometers as sb
from numpy.typing import NDArray

class HR4000CG:
    def __init__(self, device_sn:str=None) -> None:
        self.inst = inst.HR4000CG
        self.target_device_sn = device_sn
        self.device_id=""
        self.device_model=""
        self.logger = bsl_logger(self.inst)
        self.logger.info(f"Initiating bsl_instrument - SPEC({device_sn})...")
        
        self.__connect_spectrometer()
            
        if self.spec is not None:
            self.logger.device_id = self.device_id
            self.logger.success(f"READY - OceanOptics PM100D Spectrometer \"{self.device_id}\"\n\n")
        return None

    def __del__(self, *args, **kwargs) -> None:
        self.close()
        return None

    def __connect_spectrometer(self) -> None:
        """
        - Try to connect to spectrometer based on the s/n provided, if no
        s/n is provided, connect to the next available spectrometer on the bus.

        Raises:
            self.DeviceConnectionFailed: Failed to connect to spectrometer.
        """
        self.spec = None
        if len(sb.list_devices()) == 0:
            self.logger.error(f"Device not found on communication bus.\n\n\n")
            raise bsl_type.DeviceConnectionFailed
        
        self.logger.trace(f"Devices found on bus: {str(sb.list_devices())}")
        try:
            if self.target_device_sn is (None or ""):
                # with sb.Spectrometer.from_first_available() as spec_device:
                self.spec = sb.Spectrometer.from_first_available()
            elif self.target_device_sn in str(sb.list_devices()):
                # with sb.Spectrometer.from_serial_number(self.target_device_sn) as spec_device:
                self.spec = sb.Spectrometer.from_serial_number(self.target_device_sn)
            else:
                self.logger.error(f"FAILED - Device[s] found on the bus, but failed to find requested device with s/n: \"{self.target_device_sn}\".\n\n\n")
                raise bsl_type.DeviceConnectionFailed
        except:
            self.logger.error(f"FAILED - Device[s] found on the communication bus, but failed to make connection.\n\n\n")
            raise bsl_type.DeviceConnectionFailed
            
        self.device_id = self.spec.serial_number
        self.device_model = self.spec.model
        return None

    def get_wavelength(self) -> NDArray[numpy.float_]:
        """
        - wavelength array of the spectrometer
        - wavelengths in (nm) corresponding to each pixel of the spectrometer

        Returns
        -------
        wavelengths : `numpy.ndarray`
            wavelengths in (nm)
        """
        return self.spec.wavelengths()

    def get_intensity(self, correct_dark_counts: bool = False, correct_nonlinearity: bool = False) -> NDArray[numpy.float_]:
        """
        - measured intensity array in (a.u.)

        Measured intensities as numpy array returned by the spectrometer.
        The measuring behavior can be adjusted by setting the trigger mode.
        Pixels at the start and end of the array might not be optically
        active so interpret their returned measurements with care. Refer
        to the spectrometer's datasheet for further information.

        Parameters
        ----------
        correct_dark_counts : `bool`
            If requested and supported the average value of electric dark
            pixels on the ccd of the spectrometer is subtracted from the
            measurements to remove the noise floor in the measurements
            caused by non optical noise sources.
        correct_nonlinearity : `bool`
            Some spectrometers store non linearity correction coefficients
            in their eeprom. If requested and supported by the spectrometer
            the readings returned by the spectrometer will be linearized
            using the stored coefficients.

        Returns
        -------
        intensities : `numpy.ndarray`
            measured intensities in (a.u.)
        """
        return self.spec.intensities(correct_dark_counts, correct_nonlinearity)

    def get_spectrum(self, correct_dark_counts:bool=False, correct_nonlinearity:bool=False) -> NDArray[numpy.float_]:
        """
        - returns wavelengths and intensities as single array

        Uses
        ----------
        >>> (wavelengths, intensities) = spec.spectrum()

        Parameters
        ----------
        correct_dark_counts : `bool`
            see `Spectrometer.intensities`
        correct_nonlinearity : `bool`
            see `Spectrometer.intensities`

        Returns
        -------
        spectrum : `numpy.ndarray`
            combined array of wavelengths and measured intensities
        """
        return self.spec.spectrum(correct_dark_counts, correct_nonlinearity)

    def set_integration_time_micros(self, exp_us:int) -> None:
        """
        - set the integration time in microseconds

        Parameters
        ----------
        integration_time_micros : `int`
            integration time in microseconds
        """
        self.spec.integration_time_micros(exp_us)
        return None

    @property
    def integration_time_limit_us(self) -> 'tuple[int,int]':
        """
        - return the hardcoded minimum and maximum integration time

        Returns
        -------
        integration_time_micros_min_max : `tuple[int, int]`
            min and max integration time in micro seconds
        """
        return self.spec.integration_time_micros_limits
    
    @property
    def device_max_intensity(self) -> float:
        """
        - return the maximum intensity of the spectrometer

        Returns
        -------
        max_intensity : `float`
            the maximum intensity that can be returned by the spectrometer in (a.u.)
            It's possible that the spectrometer saturates already at lower values.
        """
        return self.spec.max_intensity

    @property
    def device_pixel_count(self) -> int:
        """the spectrometer's number of pixels"""
        return self.spec.pixels

    def close(self) -> None:
        try:
            if self.spec is not None:
                self.spec.close()
                del self.spec
        except:
            pass
        self.logger.success(f"CLOSED - OceanOptics PM100D Spectrometer \"{self.device_id}\"\n\n\n")
        return None




from ._inst_lib.instruments import _HR4000CG, _M69920, _PM100D, _RS_7_1, _SP_2150, _mantisCam, _DC2200, _CS260B, _Futek_USB_520
from loguru import logger as __logger
import sys as __sys

__is_logger_ready = False
__GLOBAL_LOG_LEVEL = "DEBUG"

def init_logger(LOG_LEVEL:str="DEBUG"):
    global __is_logger_ready
    global __GLOBAL_LOG_LEVEL
    __GLOBAL_LOG_LEVEL = LOG_LEVEL
    __format_str = "<cyan>{time:MM-DD at HH:mm:ss}</cyan> | <level>{level:7}</level> | {file:15}:{line:4} | <level>{message}</level>"
    __logger.remove()
    __logger.add(__sys.stdout, colorize=True, format=__format_str, level=LOG_LEVEL, diagnose=False)
    __logger.success(f"Logger initlized with LOG_LEVEL = \"{LOG_LEVEL}\".")
    __is_logger_ready = True
    return None

# Instrument Class for Thorlabs PM100D Power Meter
def PM100D(device_sn:str="") -> _PM100D.PM100D:
    if not __is_logger_ready:
        init_logger()
    return _PM100D.PM100D(device_sn)

# Instrument Class for Thorlabs DC2200 DC LED Driver
def DC2200(device_sn:str="") -> _DC2200.DC2200:
    if not __is_logger_ready:
        init_logger()
    return _DC2200.DC2200(device_sn)

# Instrument Class for NewPort Arc Lamp Supply
def M69920(device_sn:str="") -> _M69920.M69920:
    if not __is_logger_ready:
        init_logger()
    return _M69920.M69920(device_sn)

def CS260B(device_sn:str="") -> _CS260B.CS260B:
    if not __is_logger_ready:
        init_logger()
    return _CS260B.CS260B(device_sn)

# Instrument Class for OceanOptics HR4000CG Spectrometer
def HR4000CG(device_sn:str="") -> _HR4000CG.HR4000CG:
    if not __is_logger_ready:
        init_logger()
    return _HR4000CG.HR4000CG(device_sn)

# Instrument Class for Gamma Scientific RS-7-1 MultiSpectral LED Source
def RS_7_1(device_sn:str="", power_on_test:bool=True) -> _RS_7_1.RS_7_1:
    if not __is_logger_ready:
        init_logger()
    return _RS_7_1.RS_7_1(device_sn, power_on_test)

# Instrument Class for Princeton Instruments SP-2150 MonoChromator
def SP_2150(device_sn:str="") -> _SP_2150.SP_2150:
    if not __is_logger_ready:
        init_logger()
    return _SP_2150.SP_2150(device_sn)

# Instrument Class for Futek USB-520 Load Cell USB ADC
def USB_520(device_sn:str="", tear_on_startup:bool = True, reverse_negative:bool = True) -> _Futek_USB_520.USB_520:
    if not __is_logger_ready:
        init_logger()
    return _Futek_USB_520.USB_520(device_sn, tear_on_startup=tear_on_startup, reverse_negative=reverse_negative)

def mantisCam(device_sn:str="", is_GSENSE:bool = False) -> _mantisCam.MantisCamCtrl:
    if not __is_logger_ready:
        init_logger()
    return _mantisCam.MantisCamCtrl(device_sn, log_level=__GLOBAL_LOG_LEVEL, is_GSENSE = is_GSENSE)


from loguru import logger
from ..headers._bsl_inst_info import _bsl_inst_info_list
from serial.tools.list_ports import comports
import serial, subprocess, re, platform, time
from ..headers._bsl_type import _bsl_type as bsl_type

logger_opt = logger.opt(ansi=True)


# @logger_opt.catch
class _bsl_serial:
    def __init__(self, target_inst:_bsl_inst_info_list , device_sn:str="") -> None:
        logger_opt.info("    Initiating bsl_serial_service...")
        self.device_id=""
        self.inst = target_inst
        self.target_device_sn = device_sn
        self.serial_port = self._connect_serial_device()
        if self.serial_port is None:
            logger_opt.error(f"<light-blue><italic>{self.inst.MODEL} ({self.target_device_sn})</italic></light-blue> not found on serial ports.")
        pass
              
    def __del__(self) -> None:
        self.close()
        return None

    def _connect_serial_device(self) -> serial.Serial:
        if self._find_device():
            logger_opt.success(f"    {self.inst.MODEL} with DEVICE_ID: <light-blue><italic>{self.device_id}</italic></light-blue> found and connected!")
            return serial.Serial(self.serial_port_name, self.baudrate)
        return None

    def _find_device(self) -> serial.tools.list_ports_common.ListPortInfo:
    # Find first available target device by searching Serial COM ports.
    # Return serial port object.

        #Aquire all available Serial COM ports.
        com_ports_list = list(comports())
        target_port = None
        logger_opt.trace(f"    Devices found on bus:{str([port_name[0] for port_name in comports()])}")
        #Search for target device with the name of the USB device.
        for port in com_ports_list:
            temp_port = None

            if self.inst.SERIAL_SN in port[0]:
                logger_opt.info(f"    Specified device <light-blue><italic>{self.inst.MODEL}</italic></light-blue> with Serial SN <light-blue><italic>{self.inst.SERIAL_SN}</italic></light-blue> found on port <light-blue><italic>{port[0]}</italic></light-blue> by Device Serial SN search.")
                temp_port = port[0]

            if self.target_device_sn in port[0]:
                self.device_id = port[0]
                logger_opt.info(f"    Specified device <light-blue><italic>{self.inst.MODEL}</italic></light-blue> with Serial SN <light-blue><italic>{self.target_device_sn}</italic></light-blue> found on port <light-blue><italic>{port[0]}</italic></light-blue> by Device Serial SN search.")
                temp_port = port[0]
            
            if self.inst.SERIAL_NAME in port[1]:
                logger_opt.info(f"    Specified device <light-blue><italic>{self.inst.MODEL}</italic></light-blue> with Serial_Name <light-blue><italic>{self.inst.SERIAL_NAME}</italic></light-blue> found on port <light-blue><italic>{port[0]}</italic></light-blue> by Device name search.")
                temp_port = port[0]
            
            if (self.inst.USB_PID in port[2]) or (str(int(self.inst.USB_PID,16)) in port[2]):
                logger_opt.info(f"    Specified device <light-blue><italic>{self.inst.MODEL}</italic></light-blue> with USB_PID: <light-blue><italic>{self.inst.USB_PID}</italic></light-blue> found on port <light-blue><italic>{port[0]}</italic></light-blue> by USB_PID search.")
                temp_port = port[0]
            
            if temp_port is not None:
                (temp_port, baudrate) = self._check_device_resp(temp_port)
                if temp_port is not None:
                    self.baudrate = baudrate
                    self.serial_port_name = temp_port
                    return True
                continue

        if self.target_device_sn != "" and self.inst.MODEL == "USB_520":
            logger_opt.error(f"<light-blue><italic>{self.inst.MODEL} ({self.target_device_sn})</italic></light-blue> not found on serial ports.")
            raise bsl_type.DeviceConnectionFailed
        
        logger.warning(f"    No device found based on USB_PID/VID or Serial Name search!")
        
        # Failed to find device with either USB_PID or device name
        # Now try to foreach every signle serial device
        for port in com_ports_list:
            temp_port = None
            (temp_port, baudrate) = self._check_device_resp(port[0])
            if temp_port is not None:
                self.baudrate = baudrate
                self.serial_port_name = temp_port
                return True
        return None
                
    def _check_device_resp(self, temp_port) -> serial.Serial:
        # Set baudrate to common baudrates if not provided
        if self.inst.BAUDRATE != 0:
            baudrates = list([self.inst.BAUDRATE])
        else:
            baudrates = list([4800,9600,19200,28800,38400,115200])
        
        if not self.is_port_free(temp_port):
            logger_opt.warning(f"    BUSY - Device <light-blue><italic>{temp_port}</italic></light-blue> is busy, moving to next available device...")
            return None,None

        # Try to communicate with the device with each possible baudrate
        try:
            for baudrate in baudrates:
                logger_opt.info(f"    Inquiring serial port <light-blue><italic>{temp_port}</italic></light-blue> with Baudrate={baudrate}")
                # Try to open the serial port
                with serial.Serial(temp_port, baudrate, timeout=0.1) as device:
                    logger_opt.trace(f"        Connected to <light-blue><italic>{device.name}</italic></light-blue> on port <light-blue><italic>{temp_port}</italic></light-blue>")
                    
                    # If no QUERY_CMD is provided, return the port and baudrate
                    if self.inst.QUERY_CMD != "N/A":
                        # Query the device with QUERY_CMD if provided
                        device.reset_input_buffer()
                        device.write(bytes(self.inst.QUERY_CMD,'utf-8'))
                        logger_opt.trace(f"        Querry <light-blue><italic>{repr(self.inst.QUERY_CMD)}</italic></light-blue> sent to <light-blue><italic>{device.name}</italic></light-blue>")
                        time.sleep(0.5)
                    resp=""
                    try:
                        resp = repr(device.read(100).decode("utf-8")).strip('\n\r')
                    except:
                        resp = "ERROR in interpreting as UTF-8."
                    logger_opt.trace(f"        Response from <light-blue><italic>{device.name}</italic></light-blue>: {resp}")
                    
                    # Check if the response contains expected string and s/n number, if true, port found.
                    if self.inst.QUERY_E_RESP in resp:
                        if self.inst.QUERY_SN_CMD == "N/A":
                            return (temp_port, baudrate) 
                        
                        logger_opt.info(f"        <light-blue><italic>{self.inst.MODEL}</italic></light-blue> found on serial bus on port <light-blue><italic>{temp_port}</italic></light-blue>.")
                        # Check S/N to confirm matching
                        device.reset_input_buffer() 
                        device.write(bytes(self.inst.QUERY_SN_CMD,'utf-8'))
                        logger_opt.trace(f"        Querry <light-blue><italic>{repr(self.inst.QUERY_SN_CMD)}</italic></light-blue> sent to <light-blue><italic>{device.name}</italic></light-blue>")
                        time.sleep(0.5)
                        resp = (device.read(100).decode("utf-8")).strip('\n\r')
                        logger_opt.trace(f"        Response from <light-blue><italic>{device.name}</italic></light-blue>: {resp}")
                        # Use provided regular expression to extract device S/N number
                        device_id = re.search(self.inst.SN_REG, resp).group(0)
                        device.close()
                        # Return device_port and current baudrate if a positive match is confirmed
                        if self.target_device_sn in device_id:
                            self.device_id = device_id.strip('\r\n')
                            return (temp_port, baudrate) 
                        # Able to confirm device model number, but mismatch S/N number
                        logger_opt.warning(f"    S/N Mismatch - Device <light-blue><italic>{temp_port}</italic></light-blue> with S/N <light-blue><italic>{device_id}</italic></light-blue> found, not <light-blue><italic>{self.target_device_sn}</italic></light-blue> as requested, moving to next available device...")
                        break
                    device.close()
        except serial.SerialException:
            logger_opt.warning(f"    BUSY - Device <light-blue><italic>{temp_port}</italic></light-blue> is busy, moving to next available device...")
            return None,None
        return None,None

    def readline(self) -> str:
        resp = self.serial_port.readline().decode("utf-8")
        logger_opt.trace(f"        {self.inst.MODEL} - com-Serial - Resp from {self.inst.MODEL} with {repr(resp)}")
        return resp.strip('\n\r')
    
    def read(self, n_bytes:int) -> str:
        resp = self.serial_port.read(n_bytes).decode("utf-8")
        logger_opt.trace(f"        {self.inst.MODEL} - com-Serial - Resp from {self.inst.MODEL} with {repr(resp)}")
        return resp.strip('\n\r')

    def read_all(self) -> bytes:
        resp = self.serial_port.read_all()
        logger_opt.trace(f"        {self.inst.MODEL} - com-Serial - Resp from {self.inst.MODEL} with {repr(resp)}")
        return resp.strip()

    def write(self, msg:str) -> int:
        logger_opt.trace(f"        {self.inst.MODEL} - com-Serial - Write to {self.inst.MODEL} with {repr(msg)}")
        return self.serial_port.write(bytes(msg, 'utf-8'))
    
    def writeline(self, msg:str) -> int:
        msg = msg +'\r\n'
        logger_opt.trace(f"        {self.inst.MODEL} - com-Serial - Write to {self.inst.MODEL} with {repr(msg)}")
        return self.serial_port.write(bytes(msg, 'utf-8'))

    def query(self, cmd:str) -> str:
        self.flush_read_buffer()
        self.writeline(cmd)
        return self.readline()
        
    def flush_read_buffer(self) -> None:
        self.serial_port.reset_input_buffer()
        pass

    def set_serial_timeout(self, timeout:int) -> None:
        """
        Set timout in seconds for the serial communication interface.
        NOT TESTED.

        Parameters
        ----------
        timeout : `int`
            Timeout for serial communication in seconds.
        """
        self.serial_port.timeout = timeout
        pass

    def is_port_free(self, port_name):
        """
        Check if the given serial port is busy/open by another process on macOS.
        :param port_name: Full path to the device, e.g., '/dev/tty.usbserial'.
        :return: True if the port is busy, False otherwise.
        """
        if platform.system() != "Darwin":
            return True
        
        try:
            # Use the lsof command to see if the port is open by any process.
            result = subprocess.run(['lsof', port_name], capture_output=True, text=True)

            # If the output contains the port name, it's open by some process.
            if port_name in result.stdout:
                return False
            return True
        except Exception as e:
            print(f"Error: {e}")
            return True

    def close(self) -> None:
        if self.serial_port is not None:
            self.serial_port.close()
        pass
        


from loguru import logger
from ._bsl_type import _bsl_type as bsl_type
from ._bsl_inst_info import _bsl_inst_info_class as inst

logger_opt = logger.opt(ansi=True)

# @logger.catch(exclude=(bsl_type.DeviceConnectionFailed,bsl_type.DeviceInconsistentError,bsl_type.DeviceOperationError))
class _bsl_logger:
    def __init__(self, cur_inst: inst, device_id: str="N/A") -> None:
        self.__inst = cur_inst
        self.device_id = device_id
        logger.info(f"    {self.__inst.MODEL}  ({self.device_id}) - Logger instance initilized")
        pass

    def __del__(self, *args, **kwargs) -> None:
        self.close()

    def error(self, msg:str="") -> None:
        logger.error(f"ERROR - {self.__inst.MODEL}  ({self.device_id})- {msg}")
        # raise bsl_type.DeviceOperationError

    def warning(self, msg:str="") -> None:
        logger.warning(f"    {self.__inst.MODEL}  ({self.device_id}) - {msg}")

    def info(self, msg:str="") -> None:
        logger.info(f"    {self.__inst.MODEL}  ({self.device_id}) - {msg}")

    def trace(self, msg:str="") -> None:
        logger.trace(f"    {self.__inst.MODEL}  ({self.device_id}) - {msg}")

    def debug(self, msg:str="") -> None:
        logger.debug(f"    {self.__inst.MODEL}  ({self.device_id}) - {msg}")

    def success(self, msg:str="") -> None:
        logger.success(f"    {self.__inst.MODEL}  ({self.device_id}) - {msg}")

    def close(self) -> None:
        logger.info(f"    - Logger instance terminated")


from ._bsl_inst_info_class import _bsl_inst_info_class

class _bsl_inst_info_list:
    PM100D = _bsl_inst_info_class(
        MANUFACTURE="THORLAB",
        MODEL="PM100D",
        TYPE="Power Meter",
        INTERFACE="VISA",
        USB_PID="0x8078",
        USB_VID="0x1313",
        QUERY_CMD="*IDN?",
        QUERY_SN_CMD="*IDN?",
        QUERY_E_RESP="PM100D",
        SN_REG="(?<=,)P[0-9]+(?=,)"
    )

    DC2200 = _bsl_inst_info_class(
        MANUFACTURE="THORLAB",
        MODEL="DC2200",
        TYPE="LED Controller",
        INTERFACE="VISA",
        USB_PID="0x80C8",
        USB_VID="0x1313",
        QUERY_CMD="*IDN?",
        QUERY_SN_CMD="*IDN?",
        QUERY_E_RESP="DC2200",
        SN_REG="M\d{8}"
    )

    M69920 = _bsl_inst_info_class(
        MANUFACTURE="Newport Corp.",
        MODEL="M69920",
        TYPE="Power Supply",
        INTERFACE="Serial",
        BAUDRATE=9600,
        SERIAL_NAME="69920",
        QUERY_CMD="IDN?\r",
        QUERY_E_RESP="69920"
    )

    USB_520 = _bsl_inst_info_class(
        MANUFACTURE="Futek",
        MODEL="USB_520",
        TYPE="USB ADC for Load Cells",
        INTERFACE="Serial",
        BAUDRATE=9600,
        SERIAL_SN="N/A",
        QUERY_SN_CMD = "N/A",
        QUERY_E_RESP="g"
    )
    

    CS260B = _bsl_inst_info_class(
        MANUFACTURE="Newport Corp.",
        MODEL="CS260B-Q-MC-D",
        TYPE="Monochromator",
        SERIAL_NAME="???",
        INTERFACE="VISA",
        USB_PID="0x0014",
        USB_VID="0x1FDE",
        QUERY_CMD="*IDN?",
        QUERY_SN_CMD="*IDN?",
        QUERY_E_RESP="CS260B",
        SN_REG="^Newport Corp,CS260B,([^,]+)"
    )

    HR4000CG = _bsl_inst_info_class(
        MANUFACTURE="Ocean Optics",
        MODEL="HR4000CG",
        TYPE="Spectrometer",
        SERIAL_NAME="???",
        INTERFACE="USB-SDK",
        USB_PID="???",
        USB_VID="???"
    )

    TEST_DEVICE_NO_BAUD = _bsl_inst_info_class(
        MANUFACTURE="BSL",
        MODEL="TEST_DEVICE_BAUD",
        TYPE="TEST_DEVICE_BAUD",
        SERIAL_NAME="Incoming",
        INTERFACE="Serial",
        USB_PID="0x8078",
        USB_VID="0x1313",
        QUERY_CMD="*IDN?",
        QUERY_SN_CMD="*IDN?",
        QUERY_E_RESP="PM100D",
        SN_REG="(?<=,)P[0-9]+(?=,)"
    )

    TEST_DEVICE_BAUD = _bsl_inst_info_class(
        MANUFACTURE="BSL",
        MODEL="TEST_DEVICE_NO_BAUD",
        TYPE="TEST_DEVICE_BAUD",
        BAUDRATE=115200,
        SERIAL_NAME="Incoming",
        INTERFACE="Serial",
        USB_PID="0x8078",
        USB_VID="0x1313",
        QUERY_CMD="*IDN?",
        QUERY_SN_CMD="*IDN?",
        QUERY_E_RESP="PM100D",
        SN_REG="(?<=,)P[0-9]+(?=,)"
    )

    RS_7_1 = _bsl_inst_info_class(
        MANUFACTURE="Gamma Scientific",
        MODEL="RS_7_1",
        TYPE="Tunable Light Source - LED",
        BAUDRATE=460800,
        SERIAL_NAME="FT232R USB UART",
        INTERFACE="Serial",
        USB_PID="0x0403",
        USB_VID="0x6001",
        QUERY_CMD="USN\r\n",
        QUERY_E_RESP="HX0650",
        QUERY_SN_CMD="USN\r\n",
        SN_REG=".*"
    )

    SP_2150 = _bsl_inst_info_class(
        MANUFACTURE="Princeton Instruments",
        MODEL="SP_2150",
        TYPE="Monochromator",
        BAUDRATE=9600,
        SERIAL_NAME="ACTON RESEARCH CONTROLLER",
        INTERFACE="Serial",
        USB_PID="0x6001",
        USB_VID="0x0403",
        QUERY_CMD="model\r",
        QUERY_E_RESP="SP-2-150i",
        QUERY_SN_CMD="SERIAL\r",
        SN_REG="[0-9]+"
    )

    mantisCam = _bsl_inst_info_class(
        MANUFACTURE="BioSensors Lab",
        MODEL="mantisCam_Generic",
        TYPE="Camera",
        INTERFACE="ZMQ"
    )



import h5py,cv2,concurrent.futures
import numpy as np
from pathlib import Path
from loguru import logger

(GS_NIR_X, GS_NIR_Y) = (1,1)
(GS_RED_X, GS_RED_Y) = (1,2)
(GS_GREEN_X, GS_GREEN_Y) = (2,1)
(GS_BLUE_X, GS_BLUE_Y) = (2,2)

class mantis_file:
    def __init__(self, path: Path, x3_conv: bool = False, conv_param: float = 0.48):
        logger.trace(f"Init mantisCam video file {path}.")
        self.path = path
        self.x3_conv = x3_conv
        self.conv_param = conv_param

    def __getitem__(self, i) -> np.ndarray:
        if self.x3_conv:
            with h5py.File(self.path, 'r') as file:
                return self.__conv_opencv(np.array(file['camera']['frames'][i]))
        else:
            with h5py.File(self.path, 'r') as file:
                return np.array(file['camera']['frames'][i])
    

    def __conv_opencv(self, frames: np.ndarray) -> np.ndarray:
        filter = np.array([[1 +  self.conv_param, - self.conv_param]])
        if len(frames.shape) == 4:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                futures = {executor.submit(self.__process_frame_cv2, frame, filter): i for i, frame in enumerate(frames)}
            
                results = [None] * len(frames)
                for future in concurrent.futures.as_completed(futures):
                    # Retrieve original index of the frame
                    index = futures[future]
                    results[index] = future.result()
                return np.array(results)
        else:
            return self.__process_frame_cv2(frames,filter)


    def __process_frame_cv2(self, frame: np.ndarray, filter: np.ndarray) -> np.ndarray:
        # Ensure the frame is float for proper convolution. Convert it back to uint16 later.
        frame_float = np.flip(frame).astype(np.float32)
        result = cv2.filter2D(frame_float, -1, filter)
        return np.roll(np.flip(result), 1, axis=1)
    

    @property
    def file_name(self) -> str:
        return str(self.path.name)

    @property
    def system_infos(self) -> dict:
        sys_info = dict()
        with h5py.File(self.path, 'r') as file:
            for attr in file.attrs:
                sys_info[attr] = file.attrs[attr]
        return sys_info
    
    @property
    def frames(self) -> np.ndarray:
        '''
        Return all the frames inside this file with following shape:
        numpy.ndarray[#frame, #rows, #cols, #channels]
        '''
        with h5py.File(self.path, 'r') as file:
            if self.x3_conv:
                return self.__conv_opencv(np.array(file['camera']['frames']))
            else:
                return np.array(file['camera']['frames'])

    @property
    def exposure_times(self) -> np.ndarray:
        '''
        Return correspoding exposure times in us for all frames in file as a ndarray.
        '''
        with h5py.File(self.path, 'r') as file:
            return np.array(file['camera']['integration-time-expected'])
    
    @property
    def timestamps(self) -> np.ndarray:
        '''
        Return correspoding timestamps for all frames in file as a ndarray.
        '''
        with h5py.File(self.path, 'r') as file:
            return np.array(file['camera']['timestamp'])

    @property
    def frames_GS_high_gain(self) -> np.ndarray:
        with h5py.File(self.path) as file:
            return self.frames[:,:,0:self.n_cols//2,:]

    @property
    def frames_GS_low_gain(self) -> np.ndarray:
        with h5py.File(self.path) as file:
            return self.frames[:,:,self.n_cols//2:,:]

    # @property
    # def frames_GS_low_gain_NIR(self) -> np.ndarray:
    #     with h5py.File(self.path) as file:
    #         return self.frames_GS_low_gain_NIR()[:,GS_NIR_X::4,GS_NIR_Y::4,:]

    # @property
    # def frames_GS_low_gain_RGB(self) -> np.ndarray:
    #     with h5py.File(self.path) as file:
    #         return self.frames_GS_low_gain_NIR()[:,GS_NIR_X::4,GS_NIR_Y::4,:]

    
    @property
    def n_frames(self) -> int:
        with h5py.File(self.path, 'r') as file:
            return file['camera']['frames'].shape[0]
    
    @property
    def n_rows(self) -> int:
        with h5py.File(self.path, 'r') as file:
            return file['camera']['frames'].shape[1]

    @property
    def n_cols(self) -> int:
        with h5py.File(self.path, 'r') as file:
            return file['camera']['frames'].shape[2]

    @property
    def n_chans(self) -> int:
        with h5py.File(self.path, 'r') as file:
            return file['camera']['frames'].shape[3]
    
    @property
    def is_monochrome(self) -> bool:
        if self.n_chans == 1:
            return True
        return False

    @property
    def raw_data_shape(self) -> 'tuple([int,int,int,int])':
        '''
        Return the raw data shape of the file in following order:
        [#_frames, #_rows, #_cols, #_channels]
        '''
        return (self.n_frames, self.n_rows, self.n_cols, self.n_chans)
    
    @property
    def frame_shape(self) -> 'tuple([int,int,int])':
        '''
        Return the raw data shape of the file in following order:
        [#_rows, #_cols, #_channels]
        '''
        return (self.n_rows, self.n_cols, self.n_chans)
